==> luna_cow-0.5.94-py3-none-any.whl: cow/web/request.py <==
"""
Module realizes request class
"""
import email.utils
import re
import unicodedata
from collections import namedtuple
from email.message import MIMEPart
from time import perf_counter as perfTime, time
from typing import Any, Dict, Generic, Optional, Tuple, Union
from urllib.parse import unquote

import sanic
from sanic.headers import parse_content_header
from sanic.request import Request, RequestParameters

from ..constants.http_consts import MAX_URI_LENGTH_IN_LOGS
from ..errors.errors import Error, ErrorInfo
from ..errors.exception import VLException
from ..monitoring.points import DataForMonitoring
from ..utils.log import logger
from ..utils.rid import checkRequestId, generateRequestId, requestIdCtx
from .context import CONFIG_CLASS, CONTEXT_CLASS
from .utils import decodeBytes, isValidJsonRequestContentType, loadJson, loadMsgPack

# container for parsed content type
ContentType = namedtuple("ContentType", ("contentType", "charset"))
# cache for parsed content-type headers
_CONTENT_TYPE_CACHE: Dict[str, ContentType] = {}


class VLRequest(Request, Generic[CONTEXT_CLASS]):
    """
    Class for input requests.
    Args:
        logger (Logger): request logger
        requestId (str): request id
        dataForMonitoring (DataForMonitoring): data for monitoring
        startRequestTimeCheckPoint (float): start request time checkpoint (for perf count)
        startRequestTime (float): start request time
        _contentType (str): content type
        _contentCharset (Optional[str]):  content charset
        error (Optional[ErrorInfo]): error which was occurred during request processing
        streamResponse (bool): whether stream response is expected
    """

    # cached
    _cachedUriTemplates = {}
    __slots__ = (
        "requestId",
        "dataForMonitoring",
        "startRequestTimeCheckPoint",
        "startRequestTime",
        "_contentType",
        "_contentCharset",
        "error",
        "streamResponse",
    )
    logger = logger

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.startRequestTimeCheckPoint = perfTime()
        self.startRequestTime = time()
        requestId = self.headers.get("Luna-Request-Id", "")
        if not checkRequestId(requestId):
            requestId = generateRequestId(timeIsLocal=logger.logTime == "LOCAL")
        self.requestId = requestId
        requestIdCtx.set(requestId)
        self.dataForMonitoring = DataForMonitoring()
        url = self.raw_url.decode()
        if len(url) > MAX_URI_LENGTH_IN_LOGS:
            uri = f"{url[:MAX_URI_LENGTH_IN_LOGS]}..."
        else:
            uri = url
        logger.info(
            f"Begin request, method: {self.method}, uri: {uri}",
            extra={"http.request.method": self.method, "url.path": uri},
        )
        self._contentType = ""
        self._contentCharset: Optional[str] = None
        self.error: Optional[ErrorInfo] = None
        self.streamResponse = False

    def get_args(  # pylint: disable-msg=C0103
        self,
        keep_blank_values: bool = True,
        strict_parsing: bool = False,
        encoding: str = "utf-8",
        errors: str = "replace",
    ) -> RequestParameters:
        """
        Get query arguments. Overload default `keep_blank_values`
        Args:
            keep_blank_values: flag indicating whether blank values in
              percent-encoded queries should be treated as blank strings
            strict_parsing: flag indicating what to do with parsing errors
            encoding: specify how to decode percent-encoded sequences
            errors: specify how to decode percent-encoded sequences

        Returns:
            arguments
        """
        return super().get_args(
            keep_blank_values=keep_blank_values, strict_parsing=strict_parsing, encoding=encoding, errors=errors
        )

    @property
    def args(self) -> RequestParameters:
        """
        Get query arguments.

        Overload property

        Returns:
            arguments
        """
        return self.get_args()

    @property
    def config(self) -> CONFIG_CLASS:
        """
        Get application config.
        Returns:
            application config
        """
        return self.app.ctx.serviceConfig

    @property
    def json(self) -> Union[int, float, list, Dict[str, Any]]:
        """
        Overload property for getting json
        Returns:
            loaded json
        Raises:
            VLException(Error.RequestNotContainsJson): if json parsing is failed
        """

        if self.parsed_json is None:
            try:
                if self.content_type == "application/msgpack":
                    self.parsed_json = loadMsgPack(self.body)
                elif isValidJsonRequestContentType(self.content_type):
                    text = decodeBytes(body=self.body, encoding=self.charset)
                    self.parsed_json = loadJson(text)
                else:
                    raise VLException(Error.BadContentType, 400, isCriticalError=False)
                return self.parsed_json
            except ValueError as exc:
                raise VLException(Error.RequestNotContainsJson, 400, isCriticalError=False) from exc
        return self.parsed_json

    def _generateBeautyRoute(self) -> str:
        """
        Generate beauty route representation
        Returns:
            representation for monitoring
        Examples:
            `/1/patch/<arg1:{REGEXP}>/<arg2:{REGEXP}>` ->  `/patch/{arg1}/{arg2}>`
        """
        if self.uri_template:
            resource = self.uri_template
        else:
            return "/"
        params = list(self.match_info)
        for param in params:
            resource = re.sub(rf"<{param}:.+?[>]", f"{{{param}}}", resource)

        parts = resource.split("/", maxsplit=2)
        if parts[0] == "" and parts[1].isdigit() and len(parts) > 2:
            resource = "/" + parts[2]
        return resource

    @property
    def beautyRoute(self) -> str:
        """
        Get request beauty route
        Returns:
            beauty uri template
        Examples:
            `/1/patch/<arg1:{REGEXP}>/<arg2:{REGEXP}>` ->  `/patch/{arg1}/{arg2}>`
        """
        # required to check `route` due to sanic 21.3.4 `route` may equals None
        if (hasattr(self, "route") and self.route is None) or not self.uri_template:
            return "{unknown_route}"
        route = self._cachedUriTemplates.get(self.uri_template)
        if route:
            return route
        route = self._generateBeautyRoute()
        self._cachedUriTemplates[self.uri_template] = route
        return route

    def _getContentType(self) -> ContentType:
        """
        Get request content type.
        Returns:
            tuple <contentType, charset>
        """
        rawContentType = super().content_type
        cachedValue = _CONTENT_TYPE_CACHE.get(rawContentType)
        if cachedValue:
            self._contentType, self._contentCharset = cachedValue
        else:
            requestHeadersMock = MIMEPart()
            requestHeadersMock["content-type"] = rawContentType
            self._contentType = requestHeadersMock.get_content_type()
            contentDict = requestHeadersMock["content-type"].params
            self._contentCharset = contentDict.get("charset")
            cachedValue = ContentType(self._contentType, self._contentCharset)
            _CONTENT_TYPE_CACHE[rawContentType] = cachedValue
        return cachedValue

    @property
    def content_type(self) -> str:  # pylint: disable-msg=C0103
        """
        The value of content part for Content-Type HTTP header.
        Returns:
            Content-Type if header is present in request otherwise "application/octet-stream"
        """
        return self._getContentType().contentType

    @property
    def charset(self) -> Optional[str]:
        """
        The value of charset part for Content-Type HTTP header.
        Returns:
            charset specified in the Content-Type header otherwise None
        """
        return self._getContentType().charset

    @property
    def appCtx(self) -> CONTEXT_CLASS:
        """Get app context"""
        return self.app.ctx


Field = namedtuple("Field", ["type", "body", "headers"])
File = namedtuple("File", ("body", "name", "type", "headers"))


def parse_multipart_form(body: bytes, boundary: bytes) -> Tuple[RequestParameters, RequestParameters]:
    """
    Overwrites sanic `parse_multipart_form` parser since form data headers needed.
    Code/formatting is the same as the original except for the packing of the fields.
    See https://github.com/sanic-org/sanic/blob/main/sanic/request/form.py

    Parse a request body and returns fields and files
    Args:
        body: bytes request body
        boundary: bytes multipart boundary
    Returns:
        fields (RequestParameters), files (RequestParameters)
    """
    # pylint: disable-msg=C0103,C0113,R0914,R0912

    files = {}
    fields = {}

    form_parts = body.split(boundary)
    for form_part in form_parts[1:-1]:
        headers = {}
        file_name = None
        content_type = "text/plain"
        content_charset = "utf-8"
        field_name = None
        line_index = 2
        line_end_index = 0
        while not line_end_index == -1:
            line_end_index = form_part.find(b"\r\n", line_index)
            form_line = form_part[line_index:line_end_index].decode("utf-8")
            line_index = line_end_index + 2

            if not form_line:
                break

            colon_index = form_line.index(":")
            idx = colon_index + 2
            form_header_field = form_line[0:colon_index].lower()
            form_header_value, form_parameters = parse_content_header(form_line[idx:])

            if form_header_field == "content-disposition":
                field_name = form_parameters.get("name")
                file_name = form_parameters.get("filename")

                # non-ASCII filenames in RFC2231, "filename*" format
                if file_name is None and form_parameters.get("filename*"):
                    encoding, _, value = email.utils.decode_rfc2231(form_parameters["filename*"])
                    file_name = unquote(value, encoding=encoding)

                # Normalize to NFC (Apple MacOS/iOS send NFD)
                # Notes:
                # - No effect for Windows, Linux or Android clients which
                #   already send NFC
                # - Python open() is tricky (creates files in NFC no matter
                #   which form you use)
                if file_name is not None:
                    file_name = unicodedata.normalize("NFC", file_name)

            elif form_header_field == "content-type":
                content_type = form_header_value
                content_charset = form_parameters.get("charset", "utf-8")

            else:
                headers[form_header_field] = form_header_value

        if field_name:
            post_data = form_part[line_index:-4]
            if file_name is None:
                value = post_data.decode(content_charset, errors="ignore")
                from_field = Field(type=content_type, body=value, headers=headers)
                if field_name in fields:
                    fields[field_name].append(from_field)
                else:
                    fields[field_name] = [from_field]
            else:
                form_file = File(type=content_type, name=file_name, body=post_data, headers=headers)
                if field_name in files:
                    files[field_name].append(form_file)
                else:
                    files[field_name] = [form_file]
        else:
            logger.debug("Form-data field does not have a 'name' parameter in the Content-Disposition header")

    return RequestParameters(fields), RequestParameters(files)


# Overwrites sanic `parse_multipart_form` parser since form data content types needed.
sanic.request.parse_multipart_form = parse_multipart_form

==> luna_cow-0.6.82-py3-none-any.whl: cow/web/request.py <==
"""
Module realizes request class
"""

import email.utils
import re
import unicodedata
from collections import namedtuple
from email.message import MIMEPart
from time import perf_counter as perfTime, time
from urllib.parse import unquote

import sanic
from configs_core.base_service_setting import BaseServiceSettingsClass
from sanic.headers import parse_content_header
from sanic.request import Request, RequestParameters

from ..constants.http_consts import MAX_URI_LENGTH_IN_LOGS
from ..errors.errors import Error, ErrorInfo
from ..errors.exception import VLException
from ..monitoring.points import DataForMonitoring
from ..utils.log import logger
from ..utils.rid import checkRequestId, generateRequestId, requestIdCtx
from .context import LunaAppCtx
from .utils import decodeBytes, isValidJsonRequestContentType, loadJson, loadMsgPack

# container for parsed content type
ContentType = namedtuple("ContentType", ("contentType", "charset"))
# cache for parsed content-type headers
_CONTENT_TYPE_CACHE: dict[str, ContentType] = {}


class VLRequest[TCtx: LunaAppCtx](Request):
    """
    Class for input requests.
    Args:
        logger (Logger): request logger
        requestId (str): request id
        dataForMonitoring (DataForMonitoring): data for monitoring
        startRequestTimeCheckPoint (float): start request time checkpoint (for perf count)
        startRequestTime (float): start request time
        _contentType (str): content type
        _contentCharset (Optional[str]):  content charset
        error (Optional[ErrorInfo]): error which was occurred during request processing
        streamResponse (bool): whether stream response is expected
    """

    # cached
    _cachedUriTemplates = {}
    __slots__ = (
        "requestId",
        "dataForMonitoring",
        "startRequestTimeCheckPoint",
        "startRequestTime",
        "_contentType",
        "_contentCharset",
        "error",
        "streamResponse",
    )
    logger = logger

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.startRequestTimeCheckPoint = perfTime()
        self.startRequestTime = time()
        requestId = self.headers.get("Luna-Request-Id", "")
        if not checkRequestId(requestId):
            requestId = generateRequestId(timeIsLocal=logger.logTime == "LOCAL")
        self.requestId = requestId
        requestIdCtx.set(requestId)
        self.dataForMonitoring = DataForMonitoring()
        url = self.raw_url.decode()
        if len(url) > MAX_URI_LENGTH_IN_LOGS:
            uri = f"{url[:MAX_URI_LENGTH_IN_LOGS]}..."
        else:
            uri = url
        logger.info(
            f"Begin request, method: {self.method}, uri: {uri}",
            extra={"http.request.method": self.method, "url.path": uri},
        )
        self._contentType = ""
        self._contentCharset: str | None = None
        self.error: ErrorInfo | None = None
        self.streamResponse = False

    def get_args(  # pylint: disable-msg=C0103
        self,
        keep_blank_values: bool = True,
        strict_parsing: bool = False,
        encoding: str = "utf-8",
        errors: str = "replace",
    ) -> RequestParameters:
        """
        Get query arguments. Overload default `keep_blank_values`
        Args:
            keep_blank_values: flag indicating whether blank values in
              percent-encoded queries should be treated as blank strings
            strict_parsing: flag indicating what to do with parsing errors
            encoding: specify how to decode percent-encoded sequences
            errors: specify how to decode percent-encoded sequences

        Returns:
            arguments
        """
        return super().get_args(
            keep_blank_values=keep_blank_values, strict_parsing=strict_parsing, encoding=encoding, errors=errors
        )

    @property
    def args(self) -> RequestParameters:
        """
        Get query arguments.

        Overload property

        Returns:
            arguments
        """
        return self.get_args()

    @property
    def config(self) -> BaseServiceSettingsClass:
        """
        Get application config.
        Returns:
            application config
        """
        return self.app.ctx.serviceConfig

    @property
    def json(self) -> int | float | list | dict:
        """
        Overload property for getting json
        Returns:
            loaded json
        Raises:
            VLException(Error.RequestNotContainsJson): if json parsing is failed
        """

        if self.parsed_json is None:
            try:
                if self.content_type == "application/msgpack":
                    self.parsed_json = loadMsgPack(self.body)
                elif isValidJsonRequestContentType(self.content_type):
                    text = decodeBytes(body=self.body, encoding=self.charset)
                    self.parsed_json = loadJson(text)
                else:
                    raise VLException(Error.BadContentType, 400, isCriticalError=False)
                return self.parsed_json
            except ValueError as exc:
                raise VLException(Error.RequestNotContainsJson, 400, isCriticalError=False) from exc
        return self.parsed_json

    def _generateBeautyRoute(self) -> str:
        """
        Generate beauty route representation
        Returns:
            representation for monitoring
        Examples:
            `/1/patch/<arg1:{REGEXP}>/<arg2:{REGEXP}>` ->  `/patch/{arg1}/{arg2}>`
        """
        if self.uri_template:
            resource = self.uri_template
        else:
            return "/"
        params = list(self.match_info)
        for param in params:
            resource = re.sub(rf"<{param}:.+?[>]", f"{{{param}}}", resource)

        parts = resource.split("/", maxsplit=2)
        if parts[0] == "" and parts[1].isdigit() and len(parts) > 2:
            resource = "/" + parts[2]
        return resource

    @property
    def beautyRoute(self) -> str:
        """
        Get request beauty route
        Returns:
            beauty uri template
        Examples:
            `/1/patch/<arg1:{REGEXP}>/<arg2:{REGEXP}>` ->  `/patch/{arg1}/{arg2}>`
        """
        # required to check `route` due to sanic 21.3.4 `route` may equals None
        if (hasattr(self, "route") and self.route is None) or not self.uri_template:
            return "{unknown_route}"
        route = self._cachedUriTemplates.get(self.uri_template)
        if route:
            return route
        route = self._generateBeautyRoute()
        self._cachedUriTemplates[self.uri_template] = route
        return route

    def _getContentType(self) -> ContentType:
        """
        Get request content type.
        Returns:
            tuple <contentType, charset>
        """
        rawContentType = super().content_type
        cachedValue = _CONTENT_TYPE_CACHE.get(rawContentType)
        if cachedValue:
            self._contentType, self._contentCharset = cachedValue
        else:
            requestHeadersMock = MIMEPart()
            requestHeadersMock["content-type"] = rawContentType
            self._contentType = requestHeadersMock.get_content_type()
            contentDict = requestHeadersMock["content-type"].params
            self._contentCharset = contentDict.get("charset")
            cachedValue = ContentType(self._contentType, self._contentCharset)
            _CONTENT_TYPE_CACHE[rawContentType] = cachedValue
        return cachedValue

    @property
    def content_type(self) -> str:  # pylint: disable-msg=C0103
        """
        The value of content part for Content-Type HTTP header.
        Returns:
            Content-Type if header is present in request otherwise "application/octet-stream"
        """
        return self._getContentType().contentType

    @property
    def charset(self) -> str | None:
        """
        The value of charset part for Content-Type HTTP header.
        Returns:
            charset specified in the Content-Type header otherwise None
        """
        return self._getContentType().charset

    @property
    def appCtx(self) -> LunaAppCtx:
        """Get app context"""
        return self.app.ctx


Field = namedtuple("Field", ["type", "body", "headers"])
File = namedtuple("File", ("body", "name", "type", "headers"))


def parse_multipart_form(body: bytes, boundary: bytes) -> tuple[RequestParameters, RequestParameters]:
    """
    Overwrites sanic `parse_multipart_form` parser since form data headers needed.
    Code/formatting is the same as the original except for the packing of the fields.
    See https://github.com/sanic-org/sanic/blob/main/sanic/request/form.py

    Parse a request body and returns fields and files
    Args:
        body: bytes request body
        boundary: bytes multipart boundary
    Returns:
        fields (RequestParameters), files (RequestParameters)
    """
    # pylint: disable-msg=C0103,C0113,R0914,R0912

    files = {}
    fields = {}

    form_parts = body.split(boundary)
    for form_part in form_parts[1:-1]:
        headers = {}
        file_name = None
        content_type = "text/plain"
        content_charset = "utf-8"
        field_name = None
        line_index = 2
        line_end_index = 0
        while not line_end_index == -1:
            line_end_index = form_part.find(b"\r\n", line_index)
            form_line = form_part[line_index:line_end_index].decode("utf-8")
            line_index = line_end_index + 2

            if not form_line:
                break

            colon_index = form_line.index(":")
            idx = colon_index + 2
            form_header_field = form_line[0:colon_index].lower()
            form_header_value, form_parameters = parse_content_header(form_line[idx:])

            if form_header_field == "content-disposition":
                field_name = form_parameters.get("name")
                file_name = form_parameters.get("filename")

                # non-ASCII filenames in RFC2231, "filename*" format
                if file_name is None and form_parameters.get("filename*"):
                    encoding, _, value = email.utils.decode_rfc2231(form_parameters["filename*"])
                    file_name = unquote(value, encoding=encoding)

                # Normalize to NFC (Apple MacOS/iOS send NFD)
                # Notes:
                # - No effect for Windows, Linux or Android clients which
                #   already send NFC
                # - Python open() is tricky (creates files in NFC no matter
                #   which form you use)
                if file_name is not None:
                    file_name = unicodedata.normalize("NFC", file_name)

            elif form_header_field == "content-type":
                content_type = form_header_value
                content_charset = form_parameters.get("charset", "utf-8")

            else:
                headers[form_header_field] = form_header_value

        if field_name:
            post_data = form_part[line_index:-4]
            if file_name is None:
                value = post_data.decode(content_charset, errors="ignore")
                from_field = Field(type=content_type, body=value, headers=headers)
                if field_name in fields:
                    fields[field_name].append(from_field)
                else:
                    fields[field_name] = [from_field]
            else:
                form_file = File(type=content_type, name=file_name, body=post_data, headers=headers)
                if field_name in files:
                    files[field_name].append(form_file)
                else:
                    files[field_name] = [form_file]
        else:
            logger.debug("Form-data field does not have a 'name' parameter in the Content-Disposition header")

    return RequestParameters(fields), RequestParameters(files)


# Overwrites sanic `parse_multipart_form` parser since form data content types needed.
sanic.request.parse_multipart_form = parse_multipart_form
